package seco.semcocoon.reading;
/*

 ============================================================================
                   The Apache Software License, Version 1.1
 ============================================================================

 Copyright (C) 1999-2003 The Apache Software Foundation. All rights reserved.

 Redistribution and use in source and binary forms, with or without modifica-
 tion, are permitted provided that the following conditions are met:

 1. Redistributions of  source code must  retain the above copyright  notice,
    this list of conditions and the following disclaimer.

 2. Redistributions in binary form must reproduce the above copyright notice,
    this list of conditions and the following disclaimer in the documentation
    and/or other materials provided with the distribution.

 3. The end-user documentation included with the redistribution, if any, must
    include  the following  acknowledgment:  "This product includes  software
    developed  by the  Apache Software Foundation  (http://www.apache.org/)."
    Alternately, this  acknowledgment may  appear in the software itself,  if
    and wherever such third-party acknowledgments normally appear.

 4. The names "Apache Cocoon" and  "Apache Software Foundation" must  not  be
    used to  endorse or promote  products derived from  this software without
    prior written permission. For written permission, please contact
    apache@apache.org.

 5. Products  derived from this software may not  be called "Apache", nor may
    "Apache" appear  in their name,  without prior written permission  of the
    Apache Software Foundation.

 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES,
 INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 FITNESS  FOR A PARTICULAR  PURPOSE ARE  DISCLAIMED.  IN NO  EVENT SHALL  THE
 APACHE SOFTWARE  FOUNDATION  OR ITS CONTRIBUTORS  BE LIABLE FOR  ANY DIRECT,
 INDIRECT, INCIDENTAL, SPECIAL,  EXEMPLARY, OR CONSEQUENTIAL  DAMAGES (INCLU-
 DING, BUT NOT LIMITED TO, PROCUREMENT  OF SUBSTITUTE GOODS OR SERVICES; LOSS
 OF USE, DATA, OR  PROFITS; OR BUSINESS  INTERRUPTION)  HOWEVER CAUSED AND ON
 ANY  THEORY OF LIABILITY,  WHETHER  IN CONTRACT,  STRICT LIABILITY,  OR TORT
 (INCLUDING  NEGLIGENCE OR  OTHERWISE) ARISING IN  ANY WAY OUT OF THE  USE OF
 THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 This software  consists of voluntary contributions made  by many individuals
 on  behalf of the Apache Software  Foundation and was  originally created by
 Stefano Mazzocchi  <stefano@apache.org>. For more  information on the Apache
 Software Foundation, please see <http://www.apache.org/>.

*/

import java.awt.geom.AffineTransform;
import java.awt.image.AffineTransformOp;
import java.awt.image.BufferedImage;
import java.awt.image.Raster;
import java.awt.image.WritableRaster;
import java.io.IOException;
import java.io.Serializable;
import java.util.Map;

import javax.imageio.ImageIO;

import org.apache.avalon.framework.configuration.Configurable;
import org.apache.avalon.framework.configuration.Configuration;
import org.apache.avalon.framework.configuration.ConfigurationException;
import org.apache.avalon.framework.parameters.Parameters;
import org.apache.cocoon.ProcessingException;
import org.apache.cocoon.environment.SourceResolver;
import org.apache.cocoon.reading.ResourceReader;
import org.xml.sax.SAXException;

import seco.util.AnimatedGifEncoder;

import com.sun.image.codec.jpeg.ImageFormatException;
import com.sun.image.codec.jpeg.JPEGCodec;
import com.sun.image.codec.jpeg.JPEGImageEncoder;

/**
 * The <code>ImageReader</code> component is used to serve binary image data
 * in a sitemap pipeline. It makes use of HTTP Headers to determine if
 * the requested resource should be written to the <code>OutputStream</code>
 * or if it can signal that it hasn't changed.
 *
 * Parameters:
 *   <dl>
 *     <dt>&lt;width&gt;</dt>
 *     <dd>This parameter is optional. When specified it determines the width
 *         of the image that should be served.
 *     </dd>
 *     <dt>&lt;height&gt;</dt>
 *     <dd>This parameter is optional. When specified it determines the height
 *         of the image that should be served.
 *     </dd>
 *   </dl>
 *
 * @author <a href="mailto:stefano@apache.org">Stefano Mazzocchi</a>
 * @author <a href="mailto:stephan@apache.org">Stephan Michels</a>
 * @author <a href="mailto:tcurdt@apache.org">Torsten Curdt</a>
 * @version CVS $Id: SmartImageReader.java,v 1.2 2004/02/02 11:37:48 jiemakel Exp $
 */
final public class SmartImageReader extends ResourceReader implements Configurable {

    private int width;
    private int height;
    private boolean enlarge;
    private boolean retainAspect;
    private final static String ENLARGE_DEFAULT = "true";
    private final static String RETAIN_ASPECT_DEFAULT = "true";
    private String format;

	public void configure(Configuration conf) throws ConfigurationException {
		format = conf.getChild("format").getValue();
	}
	
    public void setup(SourceResolver resolver, Map objectModel, String src, Parameters par)
            throws ProcessingException, SAXException, IOException {

        super.setup(resolver, objectModel, src, par);

        width = par.getParameterAsInteger("width", 0);
        height = par.getParameterAsInteger("height", 0);

        String enlargePar = par.getParameter("allow-enlarging", ENLARGE_DEFAULT);
        if ("true".equalsIgnoreCase(enlargePar) || "yes".equalsIgnoreCase(enlargePar)){
            enlarge = true;
        } else {
            enlarge = false;
        }
        
        String retainAspectPar = par.getParameter("retain-aspect", ENLARGE_DEFAULT);
                if ("true".equalsIgnoreCase(retainAspectPar) || "yes".equalsIgnoreCase(retainAspectPar)){
                    retainAspect = true;
                } else {
                    retainAspect = false;
                }
    }

    /** 
     * Returns the affine transform that implements the scaling.
     * The behavior is the following: if both the new width and height values
     * are positive, the image is rescaled according to these new values and
     * the original aspect ration is lost.
     * Otherwise, if one of the two parameters is zero or negative, the
     * aspect ratio is maintained and the positive parameter indicates the
     * scaling.
     * If both new values are zero or negative, no scaling takes place (a unit
     * transformation is applied).
     */
    private AffineTransform getTransform(double ow, double oh, double nw, double nh) {
        double wm = 1.0d;
        double hm = 1.0d;
        if (nw > 0 && nh > 0 && retainAspect) {
            if (ow>oh) {
                wm = nw / ow;
                hm = wm;
            } else {
                hm = nh / oh;
                wm = hm;
            }
        } else {
            if (nw > 0) {
                wm = nw / ow;
                if (nh > 0) {
                    hm = nh / oh;
                } else {
                    hm = wm;
                }
            } else {
                if (nh > 0) {
                    hm = nh / oh;
                    wm = hm;
                }
            }
        }

        if (!enlarge) {
            if (retainAspect) {
                if (wm>1.0 || hm>1.0) {
                    hm=1.0;
                    wm=1.0; 
                }
            } else {
                if (wm>1.0) {
                    wm=1.0;
                }
                if (hm>1.0) {
                    hm=1.0;
                }
            }
        }
        return new AffineTransform(wm, 0.0d, 0.0d, hm, 0.0d, 0.0d);
    }

    protected void processStream() throws IOException, ProcessingException {
    	if (format!=null) {
            // since we create the image on the fly
            response.setHeader("Accept-Ranges", "none");

            BufferedImage originalImage = ImageIO.read(inputStream);
			BufferedImage scaledImage; 
			if (width > 0 || height > 0) {
				if (getLogger().isDebugEnabled()) {
					getLogger().debug("image " + ((width==0)?"?":Integer.toString(width))
									  + "x" + ((height==0)?"?":Integer.toString(height))
									  + " expires: " + expires);
				}
	            Raster original = originalImage.getRaster();
	            double ow = original.getWidth();
            	double oh = original.getHeight();
            	AffineTransformOp filter = new AffineTransformOp(getTransform(ow, oh, width, height), AffineTransformOp.TYPE_BILINEAR);
            	WritableRaster scaled = filter.createCompatibleDestRaster(original);
            	filter.filter(original, scaled);
            	scaledImage = new BufferedImage(originalImage.getColorModel(),scaled,false,null);
			} else {
				scaledImage = originalImage;
			}
			if (format.equals("jpg") || format.equals("jpeg") || format.equals("JPEG") || format.equals("JPG")) {
				try {
					JPEGImageEncoder encoder = JPEGCodec.createJPEGEncoder(out);
				    encoder.encode(scaledImage);
				} catch (ImageFormatException e) {
					System.out.println("ImageFormatException: "+e.getMessage());
				}
			} else if (format.equals("gif") || format.equals("GIF")) {
				// TODO AnimatedGifEncoder uses AWT.Grapthics2D which requires awt.headless or X-server and is probably slow..
				AnimatedGifEncoder e = new AnimatedGifEncoder();
				e.start(out);
				e.addFrame(scaledImage);
				e.finish();
			} else {
				ImageIO.write(scaledImage,format,out);
			}
			out.flush();
            inputStream.close();
        } else {
            // only read the resource - no modifications requested
            if (getLogger().isDebugEnabled()) {
                getLogger().debug("passing original resource");
            }
            super.processStream();
        }
    }

    /**
     * Generate the unique key.
     * This key must be unique inside the space of this component.
     *
     * @return The generated key consists from src and width and height
     * parameters
    */
    public Serializable getKey() {
        if (width > 0 || height > 0) {
            return this.inputSource.getURI() + ':' + this.width + ':' + this.height;
        } else {
            return super.getKey();
        }
    }
    
	/**
		 * Returns the mime-type of the resource in process.
		 */
	public String getMimeType() {
		if (format.equals("jpg") || format.equals("jpeg") || format.equals("JPEG") || format.equals("JPG")) return "image/jpeg";
		else if (format.equals("gif") || format.equals("GIF")) return "image/gif";
		else return super.getMimeType();
	}

}
